@@ -490,9 +490,11 @@ class TestArgf < Test::Unit::TestCase
           t = ""; ARGF.readpartial(1, t); s << t
         end
       rescue EOFError
+        $stdout.binmode
         puts s
       end
     SRC
+      f.binmode
       f.puts("foo")
       f.puts("bar")
       f.puts("baz")
@@ -656,16 +658,12 @@ class TestArgf < Test::Unit::TestCase
 
   def test_skip
     ruby('-e', <<-SRC, @t1.path, @t2.path, @t3.path) do |f|
-      begin
-        ARGF.skip
-      rescue
-        puts "cannot skip" # ???
-      end
+      ARGF.skip
       puts ARGF.gets
       ARGF.skip
       puts ARGF.read
     SRC
-      assert_equal("cannot skip\n1\n3\n4\n5\n6\n", f.read)
+      assert_equal("1\n3\n4\n5\n6\n", f.read)
     end
   end
 
